@@ -14,17 +14,14 @@ from kuyruk.worker import Worker
 logger = logging.getLogger(__name__)
 
 
-def main():
+def main() -> None:
     parser = argparse.ArgumentParser(conflict_handler='resolve')
 
     # Add common options
-    parser.add_argument(
-        '-v', '--version', action='version', version=__version__)
-    parser.add_argument(
-        '-a', '--app', required=True, help='path to the Kuyruk object')
+    parser.add_argument('-v', '--version', action='version', version=__version__)
+    parser.add_argument('-a', '--app', required=True, help='path to the Kuyruk object')
 
-    subparsers = parser.add_subparsers(
-        dest='subparser_name', help='sub-command name')
+    subparsers = parser.add_subparsers(dest='subparser_name', help='sub-command name')
 
     # Parser for the "worker" sub-command
     parser_worker = subparsers.add_parser('worker', help='run a worker')
@@ -76,7 +73,7 @@ def main():
     args.func(app, args)
 
 
-def run_worker(app, args):
+def run_worker(app: Kuyruk, args: argparse.Namespace) -> None:
     w = Worker(app, args)
     w.run()
 
